"use strict";
require("reflect-metadata");
const clc_1 = require("../util/clc");
class Injector {
    constructor() {
        this.temporaryValue = Symbol.for('recursive-injection');
        this.cache = new Map();
    }
    resolveInjectable(ctorFn) {
        if (!ctorFn) {
            console.error(clc_1.clc.error('Attempted to inject a falsey type.'));
            return;
        }
        if (this.cache.get(ctorFn) === this.temporaryValue) {
            console.error(clc_1.clc.error(`Recursive injection of type ${ctorFn.name || ctorFn}`));
            return;
        }
        if (this.cache.has(ctorFn)) {
            return this.cache.get(ctorFn)();
        }
        this.cache.set(ctorFn, this.temporaryValue);
        let inst = this.construct(ctorFn);
        this.cache.set(ctorFn, () => inst);
        return inst;
    }
    construct(ctorFn) {
        let types = Reflect.getOwnMetadata('design:paramtypes', ctorFn);
        if (!types)
            throw new Error(`Could not dependency inject ${ctorFn.name || ctorFn}. No design-time metadata could be reflected.`);
        let values = types.map(type => this.resolveInjectable(type));
        return new ctorFn(...values);
    }
    provide(provideMeta) {
        let ctorFn = provideMeta.provide;
        if (!ctorFn)
            throw new Error('Attempted to provide a value for a falsey type.');
        if (this.cache.has(ctorFn))
            throw new Error(`Duplicate value provided for ${ctorFn.name || ctorFn}.`);
        let tFn;
        if (this.isClassSource(provideMeta)) {
            console.log(clc_1.clc.info(`Providing ${ctorFn.name || ctorFn} using class ${provideMeta.useClass.name || provideMeta.useClass}`));
            let t = this.construct(provideMeta.useClass);
            tFn = () => t;
        }
        else if (this.isValueSource(provideMeta)) {
            console.log(clc_1.clc.info(`Providing ${ctorFn.name || ctorFn} using value ${provideMeta.useValue}`));
            let t = provideMeta.useValue;
            tFn = () => t;
        }
        else if (this.isCallbackSource(provideMeta)) {
            console.log(clc_1.clc.info(`Providing ${ctorFn.name || ctorFn} using callback ${provideMeta.useCallback.name || provideMeta.useCallback}`));
            tFn = provideMeta.useCallback;
        }
        else
            throw new Error(`Could not resolve dependency injection provider source.`);
        this.cache.set(ctorFn, tFn);
        return this;
    }
    isClassSource(meta) {
        return !!meta.useClass;
    }
    isValueSource(meta) {
        return !!meta.useValue;
    }
    isCallbackSource(meta) {
        return !!meta.useCallback;
    }
}
exports.Injector = Injector;
//# sourceMappingURL=injector.js.map