"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
require("reflect-metadata");
const _ = require("lodash");
const Sequelize = require("sequelize");
const metadata_1 = require("../metadata");
const services_1 = require("../services");
const db_impl_1 = require("./db-impl");
class OrmReflector {
    constructor(server) {
        this.server = server;
        this.models = new Map();
        let ormTransform = server.injector.resolveInjectable(services_1.OrmTransformService);
        if (!ormTransform)
            throw new Error(`Failed to resolve OrmTransformService. Can't reflect ORM models`);
        this.ormTransform = ormTransform;
    }
    get logger() {
        return this.server.logger;
    }
    init() {
        return __awaiter(this, void 0, void 0, function* () {
            let orm = this.server.meta.orm;
            if (!orm || (typeof orm.enabled !== 'undefined' && !orm.enabled) || !orm.db)
                return;
            let db = orm.db;
            let host = db.host;
            let port = undefined;
            if (typeof host !== 'string') {
                port = host.port;
                host = host.domain;
            }
            this.sql = new Sequelize(db.name, db.user, db.password, {
                host: host,
                dialect: db.dialect || 'mysql',
                port: port
            });
            this.reflectModels(this.server.meta.models || []);
            this.reflectAssociations(this.server.meta.models || []);
            yield this.sync();
        });
    }
    sync() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.sql.sync();
        });
    }
    reflectModels(models) {
        for (let q = 0; q < models.length; q++) {
            this.reflectModel(models[q]);
        }
    }
    reflectModel(modelFn) {
        if (this.models.has(modelFn))
            throw new Error(`A model was passed to the orm-reflector twice: ${modelFn.name || modelFn}.`);
        let modelProto = modelFn.prototype;
        let meta = Reflect.getOwnMetadata(metadata_1.ModelMetadataSym, modelProto);
        if (!meta)
            throw new Error(`Expecting class with @Model decorator, could not reflect model properties for ${modelProto}.`);
        meta = this.ormTransform.transformModel(meta) || meta;
        let tableName = meta.tableName || this.ormTransform.transformModelName(modelFn.name) || modelFn.name;
        let columns = {};
        let modelOptions = _.cloneDeep(meta);
        let props = Reflect.getOwnMetadata(metadata_1.ModelPropertiesSym, modelProto) || [];
        for (let q = 0; q < props.length; q++) {
            let propName = props[q];
            let propMeta = Reflect.getOwnMetadata(metadata_1.PropMetadataSym, modelProto, propName);
            if (!propMeta)
                throw new Error(`Could not find model property metadata for property ${modelFn.name || modelFn}.${propName}.`);
            propMeta = this.ormTransform.transformColumn(propMeta) || propMeta;
            let columnMeta = _.cloneDeep(propMeta);
            columnMeta.field = columnMeta.columnName || this.ormTransform.transformColumnName(propName) || propName;
            delete columnMeta.columnName;
            columns[propName] = columnMeta;
        }
        let model = this.sql.define(tableName, columns, modelOptions);
        this.models.set(modelFn, model);
    }
    reflectAssociations(models) {
        for (let q = 0; q < models.length; q++) {
            this.reflectModelAssociations(models[q]);
        }
    }
    reflectModelAssociations(modelFn) {
        let model = this.models.get(modelFn);
        if (!model)
            throw new Error(`Could not reflect model associations for a model that failed to be reflected: ${modelFn.name || modelFn}.`);
        let modelProto = modelFn.prototype;
        let meta = Reflect.getOwnMetadata(metadata_1.ModelMetadataSym, modelProto);
        if (!meta)
            throw new Error(`Expecting class with @Model decorator, could not reflect model properties for ${modelProto}.`);
        let hasMany = Reflect.getOwnMetadata(metadata_1.ModelHasManyAssociationsSym, modelProto) || [];
        for (let q = 0; q < hasMany.length; q++) {
            let propName = hasMany[q];
            let hasManyMeta = Reflect.getOwnMetadata(metadata_1.HasManyMetadataSym, modelProto, propName);
            if (!hasManyMeta)
                throw new Error(`Could not find model has-many metadata for property ${modelFn.name || modelFn}.${propName}.`);
            let foreignModelFn = hasManyMeta.foreignModel;
            let foreignModel = this.models.get(foreignModelFn);
            if (!foreignModel)
                throw new Error(`Could not create has-many association ${modelFn.name || modelFn}.${propName} to model that has not been reflected: ${foreignModelFn.name || foreignModelFn}`);
            model.hasMany(foreignModel, hasManyMeta);
        }
        let db = new db_impl_1.DbImpl(modelFn, model);
        modelFn.db = db;
    }
}
exports.OrmReflector = OrmReflector;
//# sourceMappingURL=reflector.js.map