"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
require("reflect-metadata");
require("./extend-request");
const metadata_1 = require("../metadata");
const util_1 = require("../util");
class RouterReflector {
    constructor(server, router) {
        this.server = server;
        this.router = router;
        this.controllers = {};
    }
    get logger() {
        return this.server.logger;
    }
    reflectRoutes(controllers) {
        for (var q = 0; q < controllers.length; q++) {
            this.reflectControllerRoutes(controllers[q]);
        }
    }
    reflectControllerRoutes(controllerFn) {
        if (this.controllers[controllerFn])
            throw new Error(`A controller was passed to the router-reflector twice: ${controllerFn}.`);
        var controllerInst = this.controllers[controllerFn] = this.server.injector.resolveInjectable(controllerFn);
        var controllerProto = controllerFn.prototype;
        var meta = Reflect.getOwnMetadata(metadata_1.ControllerMetadataSym, controllerProto);
        if (!meta)
            throw new Error(`Expecting class with @Controller decorator, could not reflect routes for ${controllerProto}.`);
        this.logger.info('router', `Reflecting routes for controller ${controllerFn.name}`);
        let routes = this.reflectRouteMeta(controllerProto);
        for (let q = 0; q < routes.length; q++) {
            let [routeFnName, routeMetaArr] = routes[q];
            for (var w = 0; w < routeMetaArr.length; w++) {
                let routeMeta = routeMetaArr[w];
                this.addRoute(controllerInst, routeFnName, meta, routeMeta);
            }
        }
    }
    reflectRouteMeta(controllerProto) {
        let hierarchy = util_1.inhertitanceHierarchy(controllerProto);
        let routeMeta = [];
        for (let r = 0; r < hierarchy.length; r++) {
            let fn = hierarchy[r].prototype;
            let routeNames = Reflect.getOwnMetadata(metadata_1.ControllerRoutesSym, fn) || [];
            for (var q = 0; q < routeNames.length; q++) {
                var routeFnName = routeNames[q];
                var routeMetaArr = Reflect.getOwnMetadata(metadata_1.RouteMetadataSym, fn, routeFnName) || [];
                routeMeta.push([routeFnName, routeMetaArr]);
            }
        }
        return routeMeta;
    }
    addRoute(controller, routeFnName, meta, routeMeta) {
        let policyDescriptors = [
            ...(this.server.meta.policies || []),
            ...(meta.policies || []),
            ...(routeMeta.policies || [])
        ];
        let policies = this.resolvePolicies(policyDescriptors);
        let boundRoute = controller[routeFnName].bind(controller);
        let pathPart = routeMeta.path;
        if (controller.transformPathPart)
            pathPart = controller.transformPathPart(pathPart) || pathPart;
        let fullPath = util_1.joinRoutePaths(...[
            this.server.meta.path || '',
            meta.path || '',
            pathPart
        ]);
        if (controller.transformPath)
            fullPath = controller.transformPath(fullPath) || fullPath;
        if (typeof routeMeta.method === 'undefined')
            throw new Error(`Failed to create route ${controller}.${routeFnName}. No method set!`);
        this.logger.verbose('router', `& Adding route ${routeFnName} (${routeMeta.method.toUpperCase()} ${fullPath})`);
        this.router[routeMeta.method](fullPath, this.createFullRouterFn(policies, boundRoute));
    }
    resolvePolicies(descriptors) {
        return descriptors.map((desc) => {
            let key;
            let fn;
            if (this.isPolicyCtor(desc)) {
                key = desc;
                let val = this.server.injector.resolveInjectable(desc);
                if (!val)
                    throw new Error(`Could not resolve dependency for policy: ${desc}`);
                desc = val;
            }
            if (this.isPolicyT(desc)) {
                fn = desc.handle.bind(desc);
            }
            else {
                let handler = desc;
                fn = function (req, res) {
                    return __awaiter(this, void 0, void 0, function* () {
                        yield util_1.wrapPromise(handler, req, res);
                    });
                };
            }
            return [key, fn];
        });
    }
    isPolicyCtor(desc) {
        return !this.isPolicyT(desc) && !!desc.prototype.handle;
    }
    isPolicyT(desc) {
        return !!desc.handle;
    }
    createFullRouterFn(policies, boundRoute) {
        const self = this;
        return function (req, res) {
            return __awaiter(this, void 0, void 0, function* () {
                let allResults = [];
                req.policyResults = self.createPolicyResultsFn(policies, allResults);
                let initialStatusCode = res.statusCode;
                for (var q = 0; q < policies.length; q++) {
                    let policy = policies[q];
                    let result;
                    try {
                        result = yield policy[1](req, res);
                    }
                    catch (e) {
                        this.logger.error('router', 'A policy threw an exception. Serving 500 - Internal server error');
                        this.logger.error('router', e);
                        res.status(500);
                        res.send('Internal server error');
                        return;
                    }
                    allResults.push(result);
                    if (res.statusCode !== initialStatusCode || res.headersSent)
                        return;
                }
                try {
                    yield boundRoute(req, res);
                }
                catch (e) {
                    this.logger.error('router', 'A route threw an exception. Serving 500 - Internal server error');
                    this.logger.error('router', e);
                    res.status(500);
                    res.send('Internal server error');
                    return;
                }
                if (res.statusCode === initialStatusCode && !res.headersSent) {
                    this.logger.error('router', `A route failed to send a response. Serving 404 - Not Found`);
                    res.status(404);
                    res.send(`Not found.`);
                }
            });
        };
    }
    createPolicyResultsFn(policies, allResults) {
        let keys = policies.map(poli => poli[0]);
        return function (policyFn) {
            for (var q = 0; q < keys.length; q++) {
                if (keys[q] === policyFn)
                    return allResults[q];
            }
            return undefined;
        };
    }
}
exports.RouterReflector = RouterReflector;
//# sourceMappingURL=reflector.js.map