"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments)).next());
    });
};
const express = require("express");
const bodyParser = require("body-parser");
const core_1 = require("../core");
const orm_1 = require("../orm");
const services_1 = require("../services");
const router_1 = require("../router");
const wrap_promise_1 = require("../util/wrap-promise");
var debug = require("debug")("express:server");
class Server {
    constructor(_meta) {
        this._meta = _meta;
        this.errorCode = 0;
        this._injector = new core_1.Injector();
        this._injector.provide({ provide: Server, useValue: this });
        this._injector.provide({ provide: services_1.Logger, useValue: this._logger = new services_1.Logger(this.meta.logLevel) });
        if (_meta.inject) {
            for (var q = 0; q < _meta.inject.length; q++) {
                this._injector.provide(_meta.inject[q]);
            }
        }
    }
    get logger() {
        return this._logger;
    }
    get app() {
        return this._app;
    }
    get injector() {
        return this._injector;
    }
    get meta() {
        return this._meta;
    }
    init() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                this.logger.info('miter', `Initializing miter server...`);
                yield this.createExpressApp();
                yield this.reflectOrm();
                yield this.startServices();
                this.reflectRoutes();
            }
            catch (e) {
                this.logger.error('miter', `FATAL ERROR: Failed to launch server.`);
                this.logger.error('miter', e);
                return;
            }
            this.listen();
        });
    }
    shutdown() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                this.logger.info('miter', `Shutting down miter server...`);
                yield this.stopListening();
                yield this.stopServices();
            }
            catch (e) {
                this.logger.error('miter', `FATAL ERROR: Failed to gracefully shutdown server.`);
                this.logger.error('miter', e);
            }
        });
    }
    createExpressApp() {
        this._app = express();
        this._app.use(bodyParser.urlencoded({ extended: true }), bodyParser.json());
        if (this.meta.allowCrossOrigin) {
            this.logger.warn('miter', `Warning: server starting with cross-origin policy enabled. This should not be enabled in production.`);
            this._app.use(function (req, res, next) {
                res.header("Access-Control-Allow-Origin", "*");
                res.header("Access-Control-Allow-Headers", req.header("Access-Control-Request-Headers"));
                next();
            });
        }
        if (this.meta.middleware && this.meta.middleware.length) {
            this._app.use(...this.meta.middleware);
        }
    }
    reflectOrm() {
        return __awaiter(this, void 0, void 0, function* () {
            let orm = this.meta.orm;
            if (orm && (typeof orm.enabled === 'undefined' || orm.enabled) && orm.db) {
                this.logger.trace('orm', `Initializing ORM...`);
                this.ormReflector = new orm_1.OrmReflector(this);
                yield this.ormReflector.init();
                this.logger.info('orm', `Finished initializing ORM.`);
            }
            else if (this.meta.models && this.meta.models.length) {
                this.logger.warn('orm', `Warning: Models included in server metadata, but no orm configuration defined.`);
            }
        });
    }
    startServices() {
        return __awaiter(this, void 0, void 0, function* () {
            this.logger.trace('services', `Starting services...`);
            this.serviceReflector = new services_1.ServiceReflector(this);
            yield this.serviceReflector.reflectServices(this.meta.services || []);
            this.logger.info('services', `Finished starting services.`);
        });
    }
    stopServices() {
        return __awaiter(this, void 0, void 0, function* () {
            this.logger.trace('services', `Shutting down services...`);
            yield this.serviceReflector.shutdownServices();
            this.logger.info('services', `Finished shutting down services.`);
        });
    }
    reflectRoutes() {
        this.logger.trace('router', `Loading routes...`);
        let router = express.Router();
        this.routerReflector = new router_1.RouterReflector(this, router);
        this.routerReflector.reflectRoutes(this.meta.controllers || []);
        this.app.use(router);
        this.logger.info('router', `Finished loading routes.`);
    }
    listen() {
        this.logger.info('miter', `Serving`);
        this.httpServer = this.app.listen(this.meta.port, () => this.onListening());
        this.httpServer.on("error", (err) => this.onError(err));
    }
    stopListening() {
        return __awaiter(this, void 0, void 0, function* () {
            this.logger.trace('miter', `Closing HTTP server...`);
            yield wrap_promise_1.wrapPromise((cb) => {
                if (!this.httpServer)
                    return cb();
                this.httpServer.close(cb);
            });
            this.logger.info('miter', `Finished closing HTTP server.`);
        });
    }
    onError(error) {
        if (error.syscall !== "listen") {
            throw error;
        }
        var bind = (typeof this.meta.port === "string") ? `pipe ${this.meta.port}` : `port ${this.meta.port}`;
        // handle specific listen errors with friendly messages
        switch (error.code) {
            case "EACCES":
                this.logger.error('miter', `${bind} requires elevated privileges`);
                this.errorCode = 1;
                this.httpServer = undefined;
                this.shutdown();
                break;
            case "EADDRINUSE":
                this.logger.error('miter', `${bind} is already in use`);
                this.errorCode = 1;
                this.httpServer = undefined;
                this.shutdown();
                break;
            default:
                throw error;
        }
    }
    onListening() {
        if (!this.httpServer)
            throw new Error(`onListening called, but there is no httpServer!`);
        var addr = this.httpServer.address();
        var bind = (typeof addr === "string") ? `pipe ${addr}` : `port ${addr.port}`;
        this.logger.info('miter', `Listening on ${bind}`);
    }
}
exports.Server = Server;
//# sourceMappingURL=server.js.map